## ----setup, include = FALSE---------------------------------------------------
options(timeout = Inf)
knitr::opts_chunk$set(
    message = FALSE,
    warning = FALSE,
    tidy = FALSE,
    fig.align = 'center'
)

## ----loadpackage--------------------------------------------------------------
suppressPackageStartupMessages({
    library("CleanUpRNAseq")
    #devtools::load_all("./")
    library("ggplotify")
    library("R.utils")
    library("patchwork")
})

## ----queryannotationhub-------------------------------------------------------
## load EnsDb.Hsapiens.v86
suppressPackageStartupMessages({
    library("EnsDb.Hsapiens.v86")
})
hs_ensdb_sqlite <- EnsDb.Hsapiens.v86

## ----createensdb--------------------------------------------------------------
options(timeout = max(3000, getOption("timeout")))
tmp_dir <- tempdir()
gtf_url <- paste0(
    "https://ftp.ensembl.org/pub/release-110/gtf/",
    "homo_sapiens/Homo_sapiens.GRCh38.110.gtf.gz"
)
gtf <- basename(gtf_url)
gtf <- file.path(tmp_dir, gtf)

if (!file.exists(gtf)) {
    download.file(
        url = gtf_url,
        destfile = gtf,
        mode = "wb"
    )
}

hs_ensdb_sqlite <- make_ensdb(
    gtf = gtf,
    outfile = file.path(
        tmp_dir,
        "EnsDb.hs.v110.sqlite"
    ),
    organism_latin_name = "Homo_Sapiens",
    genome_version = "GRCh38",
    Ensembl_release_version = 110
)

## ----preparesaf---------------------------------------------------------------
bam_file <- system.file("extdata", "K084CD7PCD1N.srt.bam",
    package = "CleanUpRNAseq"
)
saf_list <- get_feature_saf(
    ensdb_sqlite = hs_ensdb_sqlite,
    bamfile = bam_file,
    mitochondrial_genome = "MT"
)

## ----featurecount-------------------------------------------------------------
bams <- system.file("extdata",
    c("K084CD7PCD1N.srt.bam", "K084CD7PCD1P.srt.bam"),
    package = "CleanUpRNAseq"
)

metadata <- data.frame(
    sample_name = c("CD1AN_m2_1", "CD1AP_m2_1"),
    BAM_file = bams,
    group = c("CD1AN", "CD1AP")
)
gunzip(gtf, remove = FALSE, overwrite = TRUE)

invisible({
    capture.output({
        counts_list <- summarize_reads(
            metadata = metadata,
            isPairedEnd = TRUE,
            strandSpecific = 0,
            saf_list = saf_list,
            gtf = gsub(".gz", "", gtf),
            threads = 1,
            verbose = FALSE
        )
    })
})

## -----------------------------------------------------------------------------
## download feaureCounts results
count_url <- paste0(
    "https://zenodo.org/records/11458839/files/",
    "read_count_summary.RData?download=1"
)
download.file(
    url = count_url,
    destfile = file.path(tmp_dir, "read_count_summary.RData"),
    mode = "wb"
)
load(file.path(tmp_dir, "read_count_summary.RData"))


p_assignment_stat <-
    check_read_assignment_stat(assignment_stat = counts_summary$gtf$stat)
wrap_plots(p_assignment_stat)

## -----------------------------------------------------------------------------
metadata <- counts_summary$metadata
counts_summary$metadata <- NULL
p_read_distr <- check_read_distribution(
    featurecounts_list = counts_summary,
    metadata = metadata
)
print(p_read_distr$p)

## ----loadsalmon---------------------------------------------------------------
quant_files <- system.file("extdata",
    c("K084CD7PCD1N.quant.sf", "K084CD7PCD1P.quant.sf"),
    package = "CleanUpRNAseq"
)

metadata_toy <- data.frame(
    sample_name = c("CD1AN_m2_1", "CD1AP_m2_1"),
    salmon_quant_file = quant_files,
    group = c("CD1AN", "CD1AP")
)

salmon_counts <- salmon_res(
    metadata = metadata_toy,
    ensdb_sqlite = hs_ensdb_sqlite
)

## download the full Salmon quantification data

salmon_url <-
    paste0("https://zenodo.org/records/11458839/files/",
           "salmon_quant_summary.RData?download=1")
destfile <- file.path(tmp_dir, "salmon_quant_summary.RData")
download.file(
    url = salmon_url,
    destfile = destfile,
    mode = "wb"
)

## load the salmon_quant object
load(destfile)

## -----------------------------------------------------------------------------
check_sample_correlation(counts = salmon_quant$counts)

## -----------------------------------------------------------------------------
metadata$group <- gsub("CD1A\\(-\\)", "CD1AN", metadata$group)
metadata$group <- gsub("CD1A\\(\\+\\)", "CD1AP", metadata$group)
p_expr_distr <- check_expr_distribution(
    counts = salmon_quant$counts,
    metadata = metadata
)
wrap_plots(p_expr_distr, ncol = 1, nrow = 3)

## -----------------------------------------------------------------------------
p_expr_genes <-
    check_expressed_gene_percentage(
        metadata = metadata,
        counts = salmon_quant$counts,
        min_cpm = 1,
        abundance = salmon_quant$abundance,
        min_tpm = 1
    )
print(p_expr_genes)

## -----------------------------------------------------------------------------
## DESeq2 exploratory analysis before correction
pca_heatmap <- exploratory_analysis(
    counts = salmon_quant$counts,
    metadata = metadata
)
pca_heatmap$pca + as.ggplot(pca_heatmap$heatmap)

## ----correctcontamination-----------------------------------------------------
corrected_counts <- global_correction(
    intergenic_featureCounts_res =
        counts_summary$intergenic_region,
    salmon_res = salmon_quant
)

## -----------------------------------------------------------------------------
metadata$group <- gsub("CD1A\\(-\\)", "CD1AN", metadata$group)
metadata$group <- gsub("CD1A\\(\\+\\)", "CD1AP", metadata$group)
p_expr_distr <- check_expr_distribution(
    counts = corrected_counts,
    metadata = metadata
)
wrap_plots(p_expr_distr, ncol = 1, nrow = 3)

## ----sessionInfo, echo = FALSE------------------------------------------------
sessionInfo()

